;;; N.B. The requirements on the the a and b parameters for a
;;; value x of length L are identical to that of -cut-
;;;
;;; (A)    a's  upper range is from 1 .. L+1
;;;             lower range is from -L to 0
;;;
;;; (B)    b's  upper range is from 0 to L
;;;             lower range is from -(L+1) to -1
;;;
;;; (C) In addition we require that once converted to their upper
;;;     range values a <= b + 1.

compile_mode :pop11 +strict;

section;

define global constant procedure allbutcut( a, b, x );

    unless isinteger( a ) and isinteger( b ) do
        a or 1 -> a;
        b or -1 -> b;
        unless isinteger( a ) and isinteger( b ) do
            mishap( 'Integer (or false) arguments needed', [ ^a ^b ] )
        endunless
    endunless;

    if x.isvectorclass then

        lvars L = datalength( x );
        lvars L1 = L fi_+ 1;

        ;;; Coerce a and b to their upper ranges.  N.B. the lower range of a and b are different!
        if a fi_<= 0 then
            L1 + a -> a;
        endif;
        if b fi_< 0 then
            L1 + b -> b;
        endif;

        ;;; We need b + 1 on a few occasions below.  So we calculate it with fast
        ;;; arithmetic that MAY OVERFLOW.  But that is OK because of the requirement
        ;;; checks below.
        lvars b1 = b fi_+ 1;

        ;;; a and b have now been converted to their upper range.
        ;;; We can now check integrity.
        unless
            1 fi_<= a and a fi_<= L1 and
            0 fi_<= b and b fi_<= L and     ;;; Traps the overflow.
            a fi_<= b1
        do
            mishap( 'Indexes out of range', [ ^a ^b ^x ] )
        endunless;

        ;;; We separate out strings because -substring- exists and is efficient.
        if x.isstring then
            ;;; We separately treat cases where A or B are anchored at the
            ;;; start or end of the vectorclass.
            if a == 1 then
                if b == L then
                    nullstring
                else
                    substring( b1, L - b, x )
                endif
            elseif b == L then
                substring( 1, a fi_- 1, x )
            elseif a == b fi_+ 1 then
                x
            else
                consstring(
                    lblock
                        lvars i;
                        fast_for i from 1 to a fi_- 1 do
                            fast_subscrs( i, x )
                        endfor;
                        fast_for i from b1 to L do
                            fast_subscrs( i, x )
                        endfor;
                    endlblock,
                    a fi_+ L fi_- b1
                )
            endif;
        else
            if a == b fi_+ 1 then
                x
            else
                lvars key = x.datakey;
                if a == 1 and b == L then
                    if key == vector_key then
                        nullvector
                    else
                        class_cons( key )( 0 )
                    endif
                else
                    lvars procedure subscr = class_fast_subscr( key );
                    class_cons( key )(
                        lblock
                            lvars i;
                            fast_for i from 1 to a fi_- 1 do
                                subscr( i, x )
                            endfor;
                            fast_for i from b1 to L do
                                subscr( i, x )
                            endfor;
                        endlblock,
                        a fi_+ L fi_- b1
                    )
                endif
            endif
        endif

    elseif x.islist then

        ;;; The main problem with lists is that we want to avoid calling
        ;;; length unnecessarily (after all, it might be dynamic).  However,
        ;;; if EITHER a or b are in the LOWER range, we are bound to
        ;;; end up computing something like -length- EXCEPT for the
        ;;; values of 0/-1.  Also, when we do want to compute -length- we want
        ;;; to do it on the shortened list, for efficiency.

        if a fi_> 0 then    ;;; a is in its upper range.

            ;;; Prepare for sysconslist(_onto).
            popstackmark;

            ;;; Put the first (a-1) elements onto the stack.  Shift
            ;;; -x- down to its a'th tail.
            lvars a_1;
            fast_repeat a fi_- 1 ->> a_1 times dest( x ) -> x endrepeat;

            if b fi_>= 0 then
                ;;; upper range of b.  It is necessary to enforce the
                ;;; requirement (C) with an fi_check.  Skip over the ( b - a + 1 )
                ;;; elements.
                fast_repeat fi_check( b - a_1, 0, false ) times tl( x ) -> x endrepeat;
                sysconslist_onto( x );
            elseif b == -1 then
                sysconslist()
            else
                ;;; b is in its lower range.  L + b + 1 puts it back into the upper
                ;;; range.  However, at this point we have stripped (a-1) items off the
                ;;; front of x.  Therefore we have length( x ) = L - a + 1.  We want to
                ;;; skip ( L + b + 1 ) - a + 1 elments.  But ( L + b + 1 ) - a + 1 is
                ;;; length( x ) + b + 1.  This needs requirement (C) enforced with an
                ;;; fi_check.
                fast_repeat fi_check( listlength( x ) fi_+ b fi_+ 1, 0, false ) times tl( x ) -> x endrepeat;
                sysconslist_onto( x );
            endif

        elseif a == 0 then

            ;;; So, provided b passes requirement (C) we simply need to return the
            ;;; entire list.
            if b == -1 then
                x
            elseif b fi_< 0 then
                mishap( a, b, x, 3, 'Indexes out of range' )
            else
                ;;; b in upper range.  Must be exactly L.
                if b == listlength( x ) then
                    x
                else
                    mishap( a, b, x, 3, 'Indexes out of range' )
                endif
            endif

        else /* still working on a list but a is in the lower range and not 0 */

            ;;; Prepare for sysconslist(_onto).
            popstackmark;

            ;;; a is in its lower range.  It is coerced to its upper range.  However, it
            ;;; turns out we really want L + a = upper(a) - 1 , so that is what we calculate.
            lvars L = listlength( x );
            lvars a_1 = L fi_+ a;
            fast_repeat a_1 times dest( x ) -> x endrepeat;

            ;;; Might as well coerce -b- to its upper range since we have calculated L.
            if b fi_< 0 then
                b fi_+ L fi_+ 1 -> b;
            endif;

            ;;; Skip over the b - a + 1 elements (i.e. b - a_1).
            fast_repeat fi_check( b fi_- a_1, 0, false ) times tl( x ) -> x endrepeat;
            sysconslist_onto( x )
        endif
    elseif x.isword then
        consword( allbutcut( a, b, fast_word_string( x ) ) )
    else
        mishap( 'Invalid argument for cut', [ ^x ] )
    endif
enddefine;

endsection;

/*
;;; Test code.
lvars i, j, p, q;
for p from 1 to 6 do
    for q from p - 1 to 5 do
        for i in [% p, p - 6 %] do
            for j in [% q, q - 6 %] do
                [%
                    {% i, j %};
                    allbutcut( i, j, { a b c d e } );
                    allbutcut( i, j, [ a b c d e ] );
                    allbutcut( i, j, "abcde" );
                %] =>
            endfor
        endfor;
        nl( 1 );
    endfor
endfor;


*/
