;;; An update of cutstring.  Works on vectorclass objects, lists, and words.
;;; Also, negative indices: counting backwards from the end of the
;;; list.  The last element is -1.
;;;     allbutlast( n, x ) = cut( 1, -(n+1), x )
;;;     alllast( n, x ) = cut( -n, -1, x )
;;;     allbutfirst( n, x ) = cut( n+1, -1, x )
;;;     allfirst( n, x ) = cut( 1, n, x )
;;;     substring( n, k, x ) = cut( n, n+k, x )

;;; This rather unintelligible code is generated by a mismatch between
;;; the various data representations and efficient built-in routines.
;;; Also, the negative indices are a bit mucky.  The proper treatment
;;; is to see the indexes as having a dual representation.  Any
;;; index can be represented as a positive number (upper range) or
;;; as a negative number (lower range).

;;; N.B. The requirements on the the a and b parameters for a
;;; value x of length L are
;;;
;;; (A)    a's  upper range is from 1 .. L+1
;;;             lower range is from -L to 0
;;;
;;; (B)    b's  upper range is from 0 to L
;;;             lower range is from -(L+1) to -1
;;;
;;; (C) In addition we require that once converted to their upper
;;;     range values a <= b + 1.

compile_mode :pop11 +strict;

section;

define global constant procedure cut( a, b, x );
    unless isinteger( a ) and isinteger( b ) do
        a or 1 -> a;
        b or -1 -> b;
        unless isinteger( a ) and isinteger( b ) do
            mishap( 'Integer (or false) arguments needed', [ ^a ^b ] )
        endunless
    endunless;

    if x.isvectorclass then

        lvars L = datalength( x );
        lvars L1 = L fi_+ 1;

        ;;; Coerce a and b to their upper ranges.  N.B. the lower range of a and b are different!
        if a fi_<= 0 then
            L1 + a -> a;
        endif;
        if b fi_< 0 then
            L1 + b -> b;
        endif;

        ;;; We need b + 1 on a few occasions below.  So we calculate it with fast
        ;;; arithmetic that MAY OVERFLOW.  But that is OK because of the requirement
        ;;; checks below.
        lvars b1 = b fi_+ 1;

        ;;; a and b have now been converted to their upper range.
        ;;; We can now check integrity.
        unless
            1 fi_<= a and a fi_<= L1 and
            0 fi_<= b and b fi_<= L and     ;;; Traps the overflow.
            a fi_<= b1
        do
            mishap( 'Indexes out of range', [ ^a ^b ^x ] )
        endunless;

        ;;; We separate out strings because -substring- exists and is efficient.
        if x.isstring then
            substring( a, b1 fi_- a, x )
        else
            lvars key = x.datakey;
            lvars procedure subscr = class_fast_subscr( key );
            class_cons( key )(#|
                lvars i;
                fast_for i from a to b do
                    subscr( i, x )
                endfor
            |#)
        endif

    elseif x.islist then

        ;;; The point of this bizarre section is to avoid calling length and to call it on the
        ;;; shortest stretch when we do.
        if a fi_> 0 then

            ;;; a is in the upper range.  This iteration ensures that a <= L+1.  It transpires
            ;;; that a - 1 is a useful number so we cache it in a1.
            lvars a1;
            fast_repeat a fi_- 1 ->> a1 times tl( x ) -> x endrepeat;
            if b == -1 then
                ;;; Trivially satisfies the preconditions.
                x
            elseif b fi_>= 0 then
                ;;; upper range of b.  It is necessary to enforce the
                ;;; requirement (C) with an fi_check.
                [% fast_repeat fi_check( b - a1, 0, false ) times dest( x ) -> x endrepeat %]
            else
                ;;; b is in its lower range.  L + b + 1 puts it back into the upper
                ;;; range.  We want another ( L + b + 1 ) - a + 1 items.  However, at this
                ;;; point we have stripped (a-1) items off the front of x.  Therefore
                ;;; we have length( x ) = L - a + 1.  So ( L + b + 1 ) - a + 1 is
                ;;; length( x ) + b + 1.  This needs requirement (C) enforced with an
                ;;; fi_check.
                [% fast_repeat fi_check( length( x ) fi_+ b fi_+ 1, 0, false ) times dest( x ) -> x endrepeat %]
            endif

        else /* still working on a list but a is in the lower range */

            ;;; a is in its lower range.  It is coerced to its upper range.  However, it
            ;;; turns out we really want a-1, so that is what we calculate.
            lvars L = length( x );
            lvars a1 = L fi_+ a;
            fast_repeat a1 times tl( x ) -> x endrepeat;
            if b == -1 then
                ;;; requirement (C) is trivially enforced.
                x
            else
                ;;; Force b into its upper range.
                if b fi_< 0 then
                    L fi_+ b fi_+ 1 -> b
                endif;
                ;;; Enforce requirement (C).
                [% fast_repeat fi_check( b fi_- a1, 0, false ) times dest( x ) -> x endrepeat %]
            endif

        endif
    elseif x.isword then
        consword( cut( a, b, fast_word_string( x ) ) )
    else
        mishap( 'Invalid argument for cut', [ ^x ] )
    endif
enddefine;

endsection;

/*
;;; Test code.
lvars i, j, p, q;
for p from 1 to 4 do
    for q from max( 1, p - 1 ) to 3 do
        for i in [% p, p - 4 %] do
            for j in [% q, q - 4 %] do
                [%
                    {% i, j %};
                    cut( i, j, { a b c } );
                    cut( i, j, [ a b c ] );
                    cut( i, j, "abc" );
                %] =>
            endfor
        endfor;
        nl( 1 );
    endfor
endfor;


*/
